#include "rpi_midi_controller.h"

#include <alsa/asoundlib.h>

static snd_seq_t *seq_handle;
static int sys_port;
static int in_port;

void subscribe_system_port();

void subscribe_midi_controller(snd_seq_addr_t controller);

void handle_system_port_events(snd_seq_event_t *ev);

void handle_in_port_events(snd_seq_event_t *ev);

void setup_midi_controller() {
  if (snd_seq_open(&seq_handle, "default", SND_SEQ_OPEN_INPUT, SND_SEQ_NONBLOCK) != 0) {
    SNDERR("snd_seq_open");
  }

  if (snd_seq_set_client_name(seq_handle, "Midi Listener") != 0) {
    SNDERR("snd_seq_set_client_name");
  }

  sys_port = snd_seq_create_simple_port(seq_handle, "listen:in",
                                        SND_SEQ_PORT_CAP_WRITE | SND_SEQ_PORT_CAP_NO_EXPORT,
                                        SND_SEQ_PORT_TYPE_APPLICATION);

  if (sys_port < 0) {
    SNDERR("snd_seq_create_simple_port");
  }

  subscribe_system_port();

  in_port = snd_seq_create_simple_port(seq_handle, "listen:in",
                                       SND_SEQ_PORT_CAP_WRITE | SND_SEQ_PORT_CAP_SUBS_WRITE,
                                       SND_SEQ_PORT_TYPE_APPLICATION);

  if (in_port < 0) {
    SNDERR("snd_seq_create_simple_port");
  }

  snd_seq_addr_t controller = {.client = 24, .port = 0};
  subscribe_midi_controller(controller);
}

void execute_midi_controller() {
  snd_seq_event_t *ev = NULL;
  int ret;
  while ((ret = snd_seq_event_input(seq_handle, &ev)) > 0) {
    if (ev->dest.port == sys_port) {
      handle_system_port_events(ev);
    } else if (ev->dest.port == in_port) {
      printf("ev %02d : %#010x - %#010x - %#010x\n", ev->type, ev->data.raw32.d[0],
             ev->data.raw32.d[1], ev->data.raw32.d[2]);
    } else {
      fprintf(stderr, "unkonwn midi dest port\n");
    }
  }
  if (ret < 0 && ret != -EAGAIN) {
    SNDERR("snd_seq_event_input");
  }
}

void close_midi_controller() {}

void subscribe_system_port() {
  snd_seq_addr_t sender = {.client = 0, .port = SND_SEQ_PORT_SYSTEM_ANNOUNCE};
  snd_seq_connect_from(seq_handle, sys_port, sender.client, sender.port);
}

void subscribe_midi_controller(snd_seq_addr_t controller) {
  snd_seq_connect_from(seq_handle, in_port, controller.client, controller.port);
}

void handle_system_port_events(snd_seq_event_t *ev) {
  if (ev->type == SND_SEQ_EVENT_PORT_START) {
    snd_seq_addr_t *newport = (snd_seq_addr_t *)&ev->data;
    if (newport->client != snd_seq_client_id(seq_handle)) {
      fprintf(stdout, "New port %d:%d\n", newport->client, newport->port);
      subscribe_midi_controller(*newport);
    }
  }
}

void handle_in_port_events(snd_seq_event_t *ev) {
  if (ev->type == SND_SEQ_EVENT_PORT_START) {
    snd_seq_addr_t *newport = (snd_seq_addr_t *)&ev->data;
    if (newport->client != snd_seq_client_id(seq_handle)) {
      fprintf(stdout, "New port %d:%d\n", newport->client, newport->port);
      subscribe_midi_controller(*newport);
    }
  }
}
