#include "rpi_artnet.h"

#include <arpa/inet.h>
#include <errno.h>
#include <fcntl.h>
#include <poll.h>
#include <pthread.h>
#include <stdbool.h>
#include <stdio.h>
#include <string.h>
#include <sys/socket.h>
#include <unistd.h>

#include "rpi_artnet_utils.h"

int udpSocket = -1;
char buffer[1024];
bool isUdpListenerRunning = false;
pthread_t udpListener;
uint8_t artDmxBufferArray[16][512];

static void *artnet_udp_handler(void *arg);

void artnet_send_poll_reply(struct sockaddr_in srcAddr);

void artnet_init() {}

void start_artnet_bg_worker() {
  isUdpListenerRunning = true;

  if (pthread_create(&udpListener, NULL, artnet_udp_handler, NULL) < 0) {
    perror("pthread_create");
  }
}

void stop_artnet_bg_worker() {
  isUdpListenerRunning = false;

  if (pthread_join(udpListener, NULL) != 0) {
    perror("pthread_join");
  }
}

int artnet_get_dmx_data(unsigned int univerve, uint8_t **dmxData) {
  if (univerve > 8) {
    fprintf(stderr, "Universe %d out of bounds %d\n", univerve, 16);
    *dmxData = NULL;
    return -1;
  }
  *dmxData = artDmxBufferArray[univerve];
  return 0;
}

static void *artnet_udp_handler(void *arg) {
  struct sockaddr_in serverAddr;
  struct pollfd fds[1];
  int timeoutMs;
  int ret;
  int flags;

  /* Create UDP socket */
  udpSocket = socket(PF_INET, SOCK_DGRAM, 0);
  if (udpSocket < 0) {
    perror("Opening socket failed");
  }

  /* Set non-blocking socket */
  flags = fcntl(udpSocket, F_GETFL, 0);
  fcntl(udpSocket, F_SETFL, flags | O_NONBLOCK);

  /* pollfd structure and timeout */
  memset(fds, 0, sizeof(fds));
  fds[0].fd = udpSocket;
  fds[0].events = POLLIN;
  timeoutMs = 10;

  /* Configure settings in address struct */
  serverAddr.sin_family = AF_INET;
  serverAddr.sin_port = htons(ARTNET_PORT);
  serverAddr.sin_addr.s_addr = inet_addr("0.0.0.0");
  memset(serverAddr.sin_zero, '\0', sizeof(serverAddr.sin_zero));

  /* Bind socket with address struct */
  bind(udpSocket, (struct sockaddr *)&serverAddr, sizeof(serverAddr));

  struct sockaddr_in srcAddr;
  socklen_t srcLen = sizeof(srcAddr);

  while (isUdpListenerRunning) {

    if ((ret = poll(fds, 1, timeoutMs)) == 1) {
      ssize_t bufferLen =
          recvfrom(udpSocket, buffer, ARTNET_MAX_BUFFER, 0, (struct sockaddr *)&srcAddr, &srcLen);

      if (bufferLen <= ARTNET_MAX_BUFFER && bufferLen > sizeof(artnetHeader_t)) {
        artnetHeader_t *artnetHeader = (artnetHeader_t *)buffer;

        if (memcmp(artnetHeader->id, ARTNET_ID, sizeof(ARTNET_ID)) == 0) {
          switch (artnetHeader->opCode) {
          case OpDmx:
            if (bufferLen >= 20) {
              artDmx_t *artDmx = (artDmx_t *)buffer;
              uint16_t dmxLength = (artDmx->lengthHi << 8) | artDmx->lengthLo;
              uint8_t *artDmxBuffer = artDmxBufferArray[artDmx->subUni & 0x00ff];
              if (dmxLength <= 512) {
                // store for later use
                memcpy(artDmxBuffer, artDmx->data, dmxLength);
              }
            }
            break;

          case OpPoll:
            artnet_send_poll_reply(srcAddr);
            break;

          default:
            break;
          }
        }
      }
    } else if (ret < 0) {
      fprintf(stderr, "error polling %d: %s\n", udpSocket, strerror(errno));
    }
  }

  close(udpSocket);

  return NULL;
}

void artnet_send_poll_reply(struct sockaddr_in srcAddr) {
  /* Configure settings in address struct */
  srcAddr.sin_family = AF_INET;
  srcAddr.sin_port = htons(ARTNET_PORT);
  memset(srcAddr.sin_zero, '\0', sizeof(srcAddr.sin_zero));

  sendto(udpSocket, (uint8_t *)&artPollReply, sizeof(artPollReply_t), 0,
         (struct sockaddr *)&srcAddr, sizeof(srcAddr));
}